/*
    FEARLUS model0-5-1-5: Environment.m
    Copyright (C) 1999-2004  Macaulay Institute

    This file is part of FEARLUS model0-5-1-5, an agent-based model of land use
    change.

    FEARLUS model0-5-1-5 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-5-1-5 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the environment class.

*/

#import "Environment.h"
#import "BitString.h"
#import "LandParcel.h"
#import "LandUse.h"
#import "MiscFunc.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "EnvironmentHistory.h"
#import "ClassInfo.h"
#import "Coordinate2D.h"
#import <string.h>
#import <errno.h>
#import <objc/objc-api.h>
#import <objc/Protocol.h>

#define EOFCHECK(expr, filen, file) if(!(expr)) { \
  fprintf(stderr, "%s -- Unexpected EOF in file %s\n", sel_get_name(_cmd), \
          (filen)); \
  [(file) drop]; \
  return self; \
}

/*

getClassMatch(TypeString, OptionalSuffix, Protocol, Separator, WhichOne)

Called by getTopologyClass and getNeighbourhoodClass to return a class
matching the required type. Separator is the separator in the typeString
between the topology and the neighbourhood, and WhichOne is whether to take
the one before or after the Separator (0 for before, 1 for after).

*/

static Class getClassMatch(char *typeStr, const char *suffix, 
			   Protocol *proto, char separator, int whichOne) {
  char *type = strdup(typeStr);
  char *str;
  char *longStr;
  char *p, *q;
  Class theClass = Nil;
  id <List> classList;

  if(type == NULL) {
    perror("Memory allocation");
    exit(1);
  }
  p = strchr(type, (int)'-');
  if(p == NULL) {
    fprintf(stderr, "Invalid format in type string\n");
    return Nil;
  }

  *p = '\0';
  if(whichOne == 0) {
    q = type;		/* Set q to point to the topology class */
  }
  else {
    q = p + 1;		/* Set q to point to the neighbourhood class */
  }
  str = strdup(q);
  if(str == NULL) {
    perror("Memory allocation");
    exit(1);
  }
  longStr = (char *)xmalloc(strlen(str) + strlen(suffix) + 1);
  if(longStr == NULL) {
    perror("Memory allocation");
    exit(1);
  }

  sprintf(longStr, "%s%s", str, suffix);

  classList = [List create: scratchZone];
  [ClassInfo getClassesForProtocol: proto inList: classList];
  
  while([classList getCount] > 0) {
    Class curClass = (Class)[classList removeFirst];

    if(strcmp([curClass name], str) == 0) {
      theClass = curClass;			// Return if exact match
      break;
    }
    else if(strcmp([curClass name], longStr) == 0) {
      theClass = curClass;		/* Save in case get exact match later
					   (like some fool's going to create
					   a class called XXXTopologyTopology)
					*/
    }
  }
  if(theClass == Nil) {
    fprintf(stderr, "No class found for %s or %s%s\n", str, str, suffix);
  }
  free(type);
  free(str);
  free(longStr);
  return theClass;
}

/*

getTopologyClass(EnvironmentTypeString)

Return the Class of the topology of the environment from the environment type
string. The environment type string has the form <TOPOLOGY>-<NEIGHBOURHOOD>
(e.g. "Toroidal-Moore" or "ToroidalTopology-MooreNeighbourhood"). This function
looks in the list of classes available that follow the Topology protocol for
a class with either the same name as the TOPOLOGY bit of the string, or with
"Topology" appended onto the end of the string.

*/

static Class getTopologyClass(char *typeStr) {
  Class topologyClass = getClassMatch(typeStr, "Topology",
				      @protocol(Grid2DTopology), '-', 0);
  if(topologyClass == Nil) {
    fprintf(stderr, "Invalid environment type string: %s -- it must be of the "
	    "format <TOPOLOGY>-<NEIGHBOURHOOD>, where <TOPOLOGY> is the name "
	    "of a class following the Topology protocol, and <NEIGHBOURHOOD> "
	    "is likewise for the Neighbourhood protocol\n", typeStr);
    exit(1);
  }
  return topologyClass;
}

/*

getNeighbourhoodClass(EnvironmentTypeString)

Return the Class of the neighbourhood from the environment type string, as
for topology...

*/

static Class getNeighbourhoodClass(char *typeStr) {
  Class nbrClass = getClassMatch(typeStr, "Neighbourhood",
				 @protocol(Neighbourhood), '-', 1);

  if(nbrClass == Nil) {
    fprintf(stderr, "Invalid environment type string: %s -- it must be of the "
	    "format <TOPOLOGY>-<NEIGHBOURHOOD>, where <TOPOLOGY> is the name "
	    "of a class following the Topology protocol, and <NEIGHBOURHOOD> "
	    "is likewise for the Neighbourhood protocol\n", typeStr);
    exit(1);
  }
  return nbrClass;
}

@implementation Environment

/*

setParametersClimateBitLength:economyBitLength:landParcelBitLength:.....

Use one method to set all the model parameters that apply to the environment.
This saves overheads on calling methods. (But for all I know, probably creates
overheads because of the number of arguments.)

*/

-(void)setParameters: (Parameter *)p {
  Coordinate2D *size;

  parameter = p;
  sClimate = [parameter climateBitStringSize];
  sEconomy = [parameter economyBitStringSize];
  sLandParcel = [parameter parcelBitStringSize];
  sLandUse = [parameter useBitStringSize];
  nLandUse = [parameter nUses];
  size = [Coordinate2D create: scratchZone setX: [parameter environmentXSize]
		       Y: [parameter environmentYSize]];
  [self setSizeX: (int)size->x Y: (int)size->y];
				// N.B. This may not work, because the setSizeX
				// method is for the Creating phase only.
				// Well, it will work if this method is called
				// between createBegin and createEnd
  pLUDontCare = [parameter pUseDontCareBit];
  year = 0;
  landParcels = [List create: [self getZone]];
  landUses = [Array create: [self getZone] setCount: (int)nLandUse];
  climate = [BitString create: [self getZone] setBitCount: sClimate];
  [climate setEachBitTo: 1 withProbability: 0.5];
  economy = [BitString create: [self getZone] setBitCount: sEconomy];
  [economy setEachBitTo: 1 withProbability: 0.5];
  history = [List create: [self getZone]];
  n_centre = [Coordinate2D create: [self getZone] setX: -1 Y: -1];
  topology = [getTopologyClass([parameter environmentType])
			      create: [self getZone]
			      setSize: size
			      Radius: [parameter neighbourhoodRadius]];
  neighbourhood 
    = [getNeighbourhoodClass([parameter environmentType])
			    create: [self getZone]
			    withRadius: [parameter neighbourhoodRadius]];
  [size drop];
}

/*

setMaxColour

Set the maximum colour value that has been successfully allocated

*/

-(void)setMaxColour: (int)col {
  maxColour = col;
}

/*

calculateYield

Calculate the yield of each land parcel

*/

-(void)calculateYield {
  int x, y;

  if([arguments debugEnvironment]) {
    printf("E: Calculating yield for each land parcel\n");
  }
  for(x = 0; x < [self getSizeX]; x++) {
    for(y = 0; y < [self getSizeY]; y++) {
      [[self getObjectAtX: x Y: y] calculateYield];
    }
  }
}

/*

determineClimate

A method for computing the climate. This may either involve loading the climate
from a file, or, if not, toggling each bit of the climate with specified toggle
probabilities -- one for each bit. If the probability is 0.5 for each bit,
then effectively the climate is random, as per model0-2. Note that when the
climate object is created, it is set to random values at that time. If the
climate file does not exist, then the climate is computed using the toggle
array, as it would be if no file were specified, however, it is then saved to
the climate file.

*/

-(void)determineClimate {
  if([parameter useClimateFile]) {
    if(climateFileStatus == NOT_OPENED_YET) {
      // Initialise the climate file pointer
      if([MiscFunc fileExists: [parameter climateFile]]) {
	climateFile = fopen([parameter climateFile], "r");
	if(climateFile == NULL) {
	  fprintf(stderr, "%s -- could not open climate file %s for reading",
		  sel_get_name(_cmd), [parameter climateFile]);
	  perror("");
	  exit(1);
	}
	climateFileStatus = OPENED_FOR_READING;
      }
      else {
	climateFile = fopen([parameter climateFile], "w");
	if(climateFile == NULL) {
	  fprintf(stderr, "%s -- could not create climate file %s",
		  sel_get_name(_cmd), [parameter climateFile]);
	  perror("");
	  exit(1);
	}
	climateFileStatus = OPENED_FOR_WRITING;
      }
    }
    // Load in, or save the data. Deliberately not using an else if, because we
    // want to continue on from the initial status NOT_OPENED_YET, if that was
    // the case.
    if(climateFileStatus == OPENED_FOR_READING) {
      // We're in read mode
      char *buf;
      int i;

      buf = (char *)xmalloc([parameter climateBitStringSize] + 2);
      for(i = 0; i < [parameter climateBitStringSize]; i++) {
	buf[i] = fgetc(climateFile);
	if(!(buf[i] == '1' || buf[i] == '0')) {
	  if(feof(climateFile)) {
	    fclose(climateFile);
	    climateFileStatus = CLOSED;
	    if([arguments debugEnvironment]) {
	      printf("E: No more data from climate file %s. Will continue by "
		     "toggling bits in the climate bit string as per the "
		     "toggle array\n", [parameter climateFile]);
	    }
	    break;
	  }
	  else {
	    fprintf(stderr, "%s -- unexpected character in climate file %s. "
		    "Expecting 1 or 0, found %c\n", sel_get_name(_cmd),
		    [parameter climateFile], buf[i]);
	    exit(1);
	  }
	}
      }
      if(i == [parameter climateBitStringSize]) {
	buf[i] = '\0';
	[climate setToString: buf];
	(void)fgetc(climateFile); 	// Skip over the carriage return
      }
      if([climate getBitCount] != [parameter climateBitStringSize]) {
	fprintf(stderr, "%s -- Invalid format in climate file %s, bit string"
		" loaded has %u bits -- expecting %u\n",
		sel_get_name(_cmd), [parameter climateFile],
		[climate getBitCount], [parameter climateBitStringSize]);
	exit(1);
      }
      free(buf);
    }
    else if(climateFileStatus == OPENED_FOR_WRITING) {
      [climate toggleEachBitUsingProbs: [parameter climateToggleProbs]];
      [climate printToFile: climateFile];
      fprintf(climateFile, "\n");
      fflush(climateFile);
    }
    // Deal with the file having run out of data. Deliberately not using an
    // else if here, because the status might have been set to CLOSED during
    // reading, and no climate have been read.
    if(climateFileStatus == CLOSED) {
      [climate toggleEachBitUsingProbs: [parameter climateToggleProbs]];
    }
  }
  else {
    [climate toggleEachBitUsingProbs: [parameter climateToggleProbs]];
  }
  if([arguments debugEnvironment]) {
    printf("E: New climate: ");
    [climate print];
  }
}

/*

getClimate

An accessor method for the climate.

*/

-getClimate {
  return climate;
}

/*

determineEconomy

A method for computing the economy. Currently this is a random bit string.

*/

-(void)determineEconomy {
  if([parameter useEconomyFile]) {
    if(economyFileStatus == NOT_OPENED_YET) {
      // Initialise the economy file pointer
      if([MiscFunc fileExists: [parameter economyFile]]) {
	economyFile = fopen([parameter economyFile], "r");
	if(economyFile == NULL) {
	  fprintf(stderr, "%s -- could not open economy file %s for reading",
		  sel_get_name(_cmd), [parameter economyFile]);
	  perror("");
	  exit(1);
	}
	economyFileStatus = OPENED_FOR_READING;
      }
      else {
	economyFile = fopen([parameter economyFile], "w");
	if(economyFile == NULL) {
	  fprintf(stderr, "%s -- could not create economy file %s",
		  sel_get_name(_cmd), [parameter economyFile]);
	  perror("");
	  exit(1);
	}
	economyFileStatus = OPENED_FOR_WRITING;
      }
    }
    // Load in, or save the data. Deliberately not using an else if, because we
    // want to continue on from the initial status NOT_OPENED_YET, if that was
    // the case.
    if(economyFileStatus == OPENED_FOR_READING) {
      // We're in read mode
      char *buf;
      int i;

      buf = (char *)xmalloc([parameter economyBitStringSize] + 2);
      for(i = 0; i < [parameter economyBitStringSize]; i++) {
	buf[i] = fgetc(economyFile);
	if(!(buf[i] == '1' || buf[i] == '0')) {
	  if(feof(economyFile)) {
	    fclose(economyFile);
	    economyFileStatus = CLOSED;
	    if([arguments debugEnvironment]) {
	      printf("E: No more data from economy file %s. Will continue by "
		     "toggling bits in the economy bit string as per the "
		     "toggle array\n", [parameter economyFile]);
	    }
	    break;
	  }
	  else {
	    fprintf(stderr, "%s -- unexpected character in economy file %s. "
		    "Expecting 1 or 0, found %c\n", sel_get_name(_cmd),
		    [parameter economyFile], buf[i]);
	    exit(1);
	  }
	}
      }
      if(i == [parameter economyBitStringSize]) {
	buf[i] = '\0';
	[economy setToString: buf];
	(void)fgetc(economyFile); 	// Skip over the carriage return
      }
      if([economy getBitCount] != [parameter economyBitStringSize]) {
	fprintf(stderr, "%s -- Invalid format in economy file %s, bit string"
		" loaded has %u bits -- expecting %u\n",
		sel_get_name(_cmd), [parameter economyFile],
		[economy getBitCount], [parameter economyBitStringSize]);
	exit(1);
      }
      free(buf);
    }
    else if(economyFileStatus == OPENED_FOR_WRITING) {
      [economy toggleEachBitUsingProbs: [parameter economyToggleProbs]];
      [economy printToFile: economyFile];
      fprintf(economyFile, "\n");
      fflush(economyFile);
    }
    // Deal with the file having run out of data. Deliberately not using an
    // else if here, because the status might have been set to CLOSED during
    // reading, and no economy have been read.
    if(economyFileStatus == CLOSED) {
      [economy toggleEachBitUsingProbs: [parameter economyToggleProbs]];
    }
  }
  else {
    [economy toggleEachBitUsingProbs: [parameter economyToggleProbs]];
  }
  if([arguments debugEnvironment]) {
    printf("E: New economy: ");
    [economy print];
  }
}

/*

getEconomy

An accessor method for the economy.

*/

-getEconomy {
  return economy;
}

/*

getLandParcels

An accessor method for the list of land parcels

*/

-getLandParcels {
  return landParcels;
}

/*

createLandUses

A method for creating an array of land uses. These are all random bit strings.
Also initialise the colours. For reasons of elegance, we don't want to go
asking the observer swarm what colours are valid for land uses. That just would
not do at all. Instead, though still not elegant, we'll assume that land
uses can take any colour in the map from 40-255. Not quite sure what the proper
solution is. Is it legitimate, for example, for the colour map to be created
in the environment, and the observer swarm to ask the environment for it? It
might be, but then colour maps might be regarded as purely an observational
feature. Though the environment knows what colour things should be in the real
world... Though the observer sees these colours in their own way... Or do they?

*/

-(void)createLandUses {
  int i, colourTaken[216], r1, r2, temp, minCol, nCol, pass, j;
  id lu;

  if(maxColour < 40) {		// This is terrible!! 40 should be #defined
				// somewhere.
    minCol = 0;
    nCol = maxColour + 1;
  }
  else {
    minCol = 40;
    nCol = maxColour - 39;
  }
  if([arguments debugEnvironment]) {
    printf("E: %d colours for %u land uses\n", nCol, nLandUse);
  }
  for(pass = 1; pass < nCol / 2; pass *= 2) {}
  j = 0;
  while(pass > 0) {
    for(i = pass; i < nCol; i += pass) {
      if(i % (pass * 2) != 0) {
	colourTaken[j++] = i + minCol;
      }
    }
    pass /= 2;
                           	// Create an array of the colours to use
  }
  /* Used to randomise the colours at this point so that the colourTaken array
     had the colours in an arbitrary order. Now I don't want to do this any
     more, because it means the batch mode and gui mode runs for the same seed
     are different. This is not good. A better solution than the fixed order
     here would be to somehow use the bit-strings (match & don't care) to set
     the colour). One could then inspect from the display which land uses were
     most similar. This would mean creating the colour map at a later stage
     somehow.

     Hmmm. Unfortunately, just taking them one at a time leaves little room
     for contrast -- so you can't really see the different land uses on the
     grid. So, instead, algorithmically shuffle the pack. Put the middle one
     at the beginning, then the quartiles as the next two, and so on and so on.
     Now I just need to think of an algorithm for doing that...
  */
  for(i = 0; i < (int)nLandUse; i++) {
    lu = [LandUse create: [self getZone]];
    [lu setParameters: parameter];	// setParameters: must be called before
                                        // initialiseWithColour:environment:
    [lu initialiseWithColour: colourTaken[i % nCol] environment: self];
    [lu randomiseBitsWithDontCare: pLUDontCare];
    [lu updateParcelPortions];
    [landUses atOffset: i put: lu];
    if([arguments debugEnvironment]) {
      printf("E: Created land use %u with match string: ", [lu getPIN]);
      [[lu getMatchBitString] printToFile: stdout];
      printf(" and don't care string: ");
      [[lu getDontCareBitString] print];
    }
  }
}

/*

createLandParcels

Populate the environment with land parcels. This is done in a spiral starting
at the middle. The reason for this is so that if we use the same seed, but
a different environment size, we will get the same land parcels in the middle
(provided we do not clump).

*/

-(void)createLandParcels {
  int step;
  int dir;	// 0 for X, 1 for Y
  int sign;
  int centreX;
  int centreY;
  int x, y;
  int max;
  int nParcels;

  nParcels = [self getSizeX] * [self getSizeY];
  centreX = [self getSizeX] / 2;
  centreY = [self getSizeY] / 2;
  x = 0;
  y = 0;
  sign = 1;
  step = 1;
  dir = 0;
  max = 1;
  while(nParcels > 0) {
    if((centreX + x) >= 0 && (centreX + x) < [self getSizeX]
       && (centreY + y) >= 0 && (centreY + y) < [self getSizeY]) {
      // The spiral will build up a square. We might have a rectangular
      // environment. So only create a land parcel if it is inside the
      // environment.
      [self createLandParcelAtX: centreX + x Y: centreY + y];
      nParcels--;
    }
    if(dir == 0) {
      x += sign;
      if(x == max) {
	dir = 1;
      }
    }
    else {
      y += sign;
      if(y == max) {
	dir = 0;
	sign *= -1;
	step++;
	max = ((step / 2) * sign) + (step % 2);
      }
    }
  }
}

/*

createLandParcelAtX:Y:

Create a land parcel at the specified co-ordinate

*/

-(void)createLandParcelAtX: (int)x Y: (int)y {
  id lp;

  lp = [LandParcel create: [self getZone]];
  [lp setParameters: parameter];
  [lp setBitLength: sLandParcel];
  [lp randomiseBits];
  [lp setXPos: x YPos: y];
  [lp setEnvironment: self];
  if(maxColour >= 40) {
    [lp setMaxGraphColour: 40];
  }
  else {
    [lp setMaxGraphColour: maxColour];
  }
  [self putObject: lp atX: x Y: y];
  [landParcels addLast: lp];
  if([arguments debugEnvironment]) {
    printf("E: Created land parcel %u at (%d, %d) with bit string: ",
	   [lp getPIN], x, y);
    [[lp getAttributes] print];
  }
}

/*

getLandParcelAtX:Y:

Get the land parcel at the specified co-ordinate. Implement a wrap-around
effect (so bounded environments beware...)

*/

-(LandParcel *)getLandParcelAtX: (int)x Y: (int)y {
  if(x < 0 || x >= [self getSizeX]) {
    x = [MiscFunc Mod: x inRange0To: [self getSizeX]];
  }
  if(y < 0 || y >= [self getSizeY]) {
    y = [MiscFunc Mod: y inRange0To: [self getSizeY]];
  }
  return (LandParcel *)[self getObjectAtX: x Y: y];
}

/*

clumpLandParcels

Use the parameter clump object to clump the land parcels

*/

-(void)clumpLandParcels {
  int x, y;

  if([parameter clumper] != nil) {
    [[parameter clumper] clumpEnvironment: self];
    if([arguments debugEnvironment]) {
      printf("E: New bit strings for land parcels:\n");
      for(x = 0; x < [self getSizeX]; x++) {
	for(y = 0; y < [self getSizeY]; y++) {
	  LandParcel *lp;

	  lp = [self getObjectAtX: x Y: y];
	  printf("\t%u at (%d, %d): ", [lp getPIN], x, y);
	  [[lp getAttributes] print];
	}
      }
    }
  }
}

/*

printTextParcelBitstrings

A method to print the parcel bit strings as text to standard output. Loops 
through each bit and writes an ASCII representation of the grid (one char per
cell) showing the setting of that bit.

*/

-(void)printTextParcelBitstrings {
  unsigned a;

  for(a = 0; a < [parameter parcelBitStringSize]; a++) {
    [self printTextParcelBitstringsBit: a];
  }
}

/*

printTextParcelBitstringsBit:

Print a grid showing the bitstring values for a single bit of the Land Parcel
bit strings

*/

-(void)printTextParcelBitstringsBit: (unsigned)a {
  int x, y;

  printf("E: %s: Bit %u:\n", sel_get_name(_cmd), a);
  for(y = [self getSizeY] - 1; y >= 0; y--) {
    printf("\t   %d ", y % 10);
    for(x = 0; x < [self getSizeX]; x++) {
      if([[self getObjectAtX: x Y: y] getOneAttribute: a] != 0) {
	printf("X");
      }
      else {
	printf(".");
      }
    }
    printf("\n");
    if(y == 0) {
      printf("\ty\n\t^    ");
      for(x = 0; x < [self getSizeX]; x++) {
	printf("%d", x % 10);
      }
      printf("\n");
    }	
  }
  printf("\t|\n\t+->x\t('X'->1, '.'->0)\n\n");
}

/*

getNLandUses

Return the number of Land Uses

*/

-(unsigned)getNLandUse {
  return nLandUse;
}

/*

getLandUses

Return the array of valid land uses

*/

-getLandUses {
  return(landUses);
}

/*

getNextNeighbourOfX:Y:

Return the object stored at the next neighbourhood location of the object at
(X, Y). This provides a means for looping through the neighbours without
worrying about how neighbourhood is defined.

The default neighbourhood function in the Environment class is Von-Neumann
toroidal. This is implemented differently here than in model 0-2. In model 0-2,
if the neighbourhood radius is large enough in comparison to the size of the
grid, it is possible for neighbours to be repeated and for a cell to see itself
as a neighbour. Specifically, if d is the neighbourhood radius, and n the size
of the grid, then if 2d >= n, you could get neighbours being repeated, and
be your own neighbour. Consider the following 3x3 grid:

                      +-----+-----+-----+
                      |     |     |     |
                      |  A  |  B  |  C  |
                      |     |     |     |
                      +-----+-----+-----+
                      |     |     |     |
                      |  D  |  E  |  F  |
                      |     |     |     |
                      +-----+-----+-----+
                      |     |     |     |
                      |  G  |  H  |  J  |
                      |     |     |     |
                      +-----+-----+-----+

Consider the neighbours of E if d == 2. With a toroidal treatment, the
neighbours to the north are B and then H (by wrap-around). To the South, H then
B. Similarly, east and west are F and D. Thus the set of neighbours of E
contain B, D, F and H uniquely, but each will be considered twice. If d == 3,
then you will get B, D, F and H repeated twice, with 4 repetitions of E. E now
sees itself as a neighbour with twice the strength of any one of its other
neighbours. Not good when you do not want E to see itself as a neighbour in the
first place!

Workaround:

If Environment X Size (N) <= 2 * neighbourhood radius, then search from 0 to N
on the X axis.
If Environment Y Size (M) <= 2 * neighbourhood radius, then search from 0 to M
on the Y axis.

This means neighbours will be returned *in a different order* than they would
be if we searched from x,y - neighbourhood radius to x,y + neighbourhood radius
-- though it would be a poor simulation that was biased in the order neighbours
are considered in any case.

*/

-getNextNeighbourOfX: (int) x Y: (int) y {
  id <Grid2DSpatial> search;

  if((x != n_centre->x) || (y != n_centre->y) || ![topology continueSearch]) {
    // Initialisation
    n_centre->x = x;
    n_centre->y = y;

    search = [topology startSearchWithCentre: n_centre];
  }
  else {
    // Continuation
    search = [topology nextSearchItem];
    if(![topology continueSearch]) return nil;
  }
  // Loop to find a valid neighbourhood co-ordinate
  while((([MiscFunc Mod: [search getX] inRange0To: [self getSizeX]] == x)
	 && ([MiscFunc Mod: [search getY] inRange0To: [self getSizeY]] == y))
				// Not a neighbour of yourself
	|| !([neighbourhood areNeighbours: n_centre and: search])) {
    search = [topology nextSearchItem];
    if(![topology continueSearch]) return nil;
  }
  return([self getObjectAtX: [MiscFunc Mod: [search getX]
				       inRange0To: [self getSizeX]]
	       Y: [MiscFunc Mod: [search getY] inRange0To: [self getSizeY]]]);
}

/*

-oldgetNextNeighbourOfX: (int) x Y: (int) y {
  if((x != n_x) || (y != n_y) || n_restart) {
    // Initialisation
    if(2 * NEIGHBOURHOODRADIUS >= [self getSizeX]) {
      n_minX = 0;
      n_maxX = [self getSizeX] - 1;
    }
    else {
      n_minX = (x - NEIGHBOURHOODRADIUS);
      n_maxX = (x + NEIGHBOURHOODRADIUS);
    }
    if(2 * NEIGHBOURHOODRADIUS >= [self getSizeY]) {
      n_minY = 0;
      n_maxY = [self getSizeY] - 1;
    }
    else {
      n_minY = (y - NEIGHBOURHOODRADIUS);
      n_maxY = (y + NEIGHBOURHOODRADIUS);
    }
    n_lastX = n_minX;
    n_lastY = n_minY;
    n_x = x;
    n_y = y;
    n_restart = 0;
  }
  else {
    // Continuation
    n_lastX++;
    if(n_lastX > n_maxX) {
      n_lastX = n_minX;
      n_lastY++;
    }	
    if(n_lastY > n_maxY) {
      n_restart = 1;
      return(nil);
    }
  }
  // Loop to find a valid neighbourhood co-ordinate
  while((([MiscFunc Mod: n_lastX inRange0To: [self getSizeX]] == x)
	 && ([MiscFunc Mod: n_lastY inRange0To: [self getSizeY]] == y))
				// Not a neighbour of yourself
	|| (([MiscFunc Mod: n_lastX inRange0To: [self getSizeX]] != x)
	    && ([MiscFunc Mod: n_lastY inRange0To: [self getSizeY]] != y))) {
				// Exclude non-von Neumann neighbours
    n_lastX++;
    if(n_lastX > n_maxX) {
      n_lastX = n_minX;
      n_lastY++;
    }	
    if(n_lastY > n_maxY) {
      n_restart = 1;
      return(nil);
    }
  }
  if([arguments debugInnerCoreMethods]) {
    printf("E: %s: Next von-neumann toroidal neighbour of (%d, %d) = "
	   "(%d, %d)\n", sel_get_name(_cmd), x, y,
	   [MiscFunc Mod: n_lastX inRange0To: [self getSizeX]],
	   [MiscFunc Mod: n_lastY inRange0To: [self getSizeY]]);
  }
  return([self getObjectAtX: [MiscFunc Mod: n_lastX
				       inRange0To: [self getSizeX]]
	       Y: [MiscFunc Mod: n_lastY inRange0To: [self getSizeY]]]);
}

*/

/*

getEconomySize

Return the size of the economy bitstring

*/

-(unsigned)getEconomyBitLength {
  return(sEconomy);
}

/*

getClimateSize

Return the size of the climate bitstring

*/

-(unsigned)getClimateBitLength {
  return(sClimate);
}

/*

getLandParcelSize

Return the size of the land parcel bitstring

*/

-(unsigned)getLandParcelBitLength {
  return(sLandParcel);
}

/*

getLandUseSize

Return the size of the land use bitstring

*/

-(unsigned)getLandUseBitLength {
  return(sLandUse);
}

/*

getTopology

Return the topology object

*/

-(id <Grid2DTopology>)getTopology {
  return topology;
}

/*

getNeighbourhoodFunction

Return the neighbourhood function object

*/

-(id <Neighbourhood>)getNeighbourhoodFunction {
  return neighbourhood;
}

/*

newYear

actions to perform in the environment at the beginning of a new year / end
of an old year:

- Make sure the histories of the land parcels and land uses are updated
- Increment the year

*/

-(void)newYear {
  int xx, yy;
  LandParcel *lp;

  year++;
  printf("--------------------------------------------------------------------"
	 "----------\n");
  printf("New year: %u\n", year);
}

/*

getYear

Return the current year

*/

-(unsigned)getYear {
  return(year);
}

/*

rankLandParcels

Rank the land parcels. Send a message to each land parcel to tell it its rank
in the population of (a) number of changes of land use and (b) number of
changes of land manager. The algorithm is simple. Loop through the land parcels
to find the maximum and minimum value of the variable of interest. Then loop
through them again to set their rank, which is given by:

                   my_value_for_the_variable - minimum
	    1.0 -  -----------------------------------
		            maximum - minimum

It is important that the result is given as a number between 0 and 1, so that
land parcels do not need to know about information like the total number of
land parcels. They need to be able to turn this number straight into a grey
level.

*/

-(void)rankLandParcels {
  id iterator, lp;
  double minLU, minLMgr, maxLU, maxLMgr, myLU, myLMgr, diffLU, diffLMgr;
  int firstTime;

  for(iterator = [landParcels begin: scratchZone],
	[iterator next], firstTime = 0;
      [iterator getLoc] == Member;
      [iterator next]) {
    lp = [iterator get];
    if(firstTime) {
      minLU = maxLU = (double)[lp getNLUChange];
      minLMgr = maxLMgr = (double)[lp getNLMgrChange];
    }
    else {
      myLU = (double)[lp getNLUChange];
      myLMgr = (double)[lp getNLMgrChange];
      minLU = (myLU < minLU) ? myLU : minLU;
      maxLU = (myLU > maxLU) ? myLU : maxLU;
      minLMgr = (myLMgr < minLMgr) ? myLMgr : minLMgr;
      maxLMgr = (myLMgr > maxLMgr) ? myLMgr : maxLMgr;
    }
  }
  diffLU = maxLU - minLU;
  diffLMgr = maxLMgr - minLMgr;
  // At this point, the iterator index should be at the end of the list
  // -- at the special non-member item at the end of the list, in fact,
  // so now iterate back through the list, beginning with a move to the
  // last member of the list
  for([iterator prev]; [iterator getLoc] == Member; [iterator prev]) {
    lp = [iterator get];
    myLU = (double)[lp getNLUChange];
    myLMgr = (double)[lp getNLMgrChange];
    [lp setChangeLURank: (1.0 - ((myLU - minLU) / diffLU))];
    [lp setChangeLMgrRank: (1.0 - ((myLMgr - minLMgr) / diffLMgr))];
  }
  [iterator drop];
}

/*

recordHistory

Record the history

*/

-(void)recordHistory {
  id iterator;
  LandParcel *lp;
  LandUse *lu;
  EnvironmentHistory *newHd;

  newHd = [EnvironmentHistory create: [self getZone]];
  [newHd initialiseWithYear: year climate: climate economy: economy];
  [history addFirst: newHd];
  if([parameter dumpHistory]
     && [parameter maxHistoryLength] < [history getCount]) {
    [[history removeLast] drop];
  }
  for(iterator = [landParcels begin: scratchZone], [iterator next];
      [iterator getLoc] == Member;
      [iterator next]) {
    lp = [iterator get];
    [lp recordHistory];
  }
  [iterator drop];
  for(iterator = [landUses begin: scratchZone], [iterator next];
      [iterator getLoc] == Member;
      [iterator next]) {
    lu = [iterator get];
    [lu recordHistory];
  }
  [iterator drop];
}


/*

writeHistory

Tell all the land parcels and land uses to write their histories to a file.

*/

-(void)writeHistory: (FILE *)fp {
  fprintf(fp, "Year\tClimate\tEconomy%s", [arguments nl]);
  [history forEach: M(printToFile:) : (id)fp];
  fprintf(fp, "Type\tID\tYear\tYield\tLand Use\tLand Manager\tStrategyMethod\t"
	  "X\tY\tAttributes\t%s", [arguments nl]);
  [landParcels forEach: M(printHistoryToFile:) : (id)fp];
  fprintf(fp, "Type\tID\tYear\tNo of Parcels Assigned\tMatch String\t"
	  "Don't Care String%s", [arguments nl]);
  [landUses forEach: M(writeLandUseHistoryToFile:) : (id)fp];
}

/*

saveLandParcelsToFileNamed:

Save the land parcels to the file. Actually, a file called name.lp will be
created if the name doesn't have a .lp suffix.

File format:

EnvironmentXSize: xxx
EnvironmentYSize: yyy
LandParcelBitStringLength: nnn
XCoord YCoord BitString
xx0	yy0	bb0
xx1	yy1	bb1
...

*/

-saveLandParcelsToFileNamed: (char *)fname {
  id file;
  char *realfname;
  char *buf;
  int x, y;

  realfname = [MiscFunc getUsableFileNameFrom: fname withSuffix: ".lp"];
  if(realfname == NULL) {
    fprintf(stderr, "%s -- could not create datafile\n", sel_get_name(_cmd));
    return self;
  }
  file = [OutFile create: [self getZone] withName: realfname];
  if(file == nil) {
    fprintf(stderr, "%s -- could not create file ", sel_get_name(_cmd));
    perror(realfname);
    free(realfname);
    return self;
  }
  buf = xmalloc([parameter parcelBitStringSize] + 2);
  // 2 is for '\0' at the end and one for luck
  [file putString: "EnvironmentXSize:\t"];
  [file putInt: [self getSizeX]];
  [file putNewLine];
  [file putString: "EnvironmentYSize:\t"];
  [file putInt: [self getSizeY]];
  [file putNewLine];
  [file putString: "LandParcelBitStringLength:\t"];
  [file putUnsigned: [parameter parcelBitStringSize]];
  [file putNewLine];
  [file putString: "XCoord\tYCoord\tBitString\n"];
  for(x = 0; x < [self getSizeX]; x++) {
    for(y = 0; y < [self getSizeY]; y++) {
      [file putInt: x];
      [file putTab];
      [file putInt: y];
      [file putTab];
      if([parameter parcelBitStringSize] > 0) {
	[[[self getObjectAtX: x Y: y] getAttributes] printToString: buf];
	[file putString: buf];
      }
      else {
	[file putString: "E"];
      }
      [file putNewLine];
    }
  }
  [file drop];
  free(realfname);
  free(buf);
  return self;
}

/*

loadLandParcelsFromFileNamed:

Load the land parcels from the file. See saveLandParcelsToFileNamed: for a
description of the format. Now, this assumes that land parcels have been
created (so, in fact, the environment can allow you to load in land parcels
during the simulation), so don't forget. Each object read must be tediously
checked. Even then, because of laziness, I still haven't checked for end of 
file with each read, but if I did the number of lines of code to read in a
single symbol would start to get silly. But there may well be unexpected
seg faults. Oh bugger I'll do it anyway.

*/

-loadLandParcelsFromFileNamed: (char *)fname {
  id file;
  char *buf;
  char word[100];
  int x, y;
  int dummyXSize, dummyYSize, dummyX, dummyY;
  unsigned dummyBSLength;
  LandParcel *lp;

  if([arguments debugEnvironment]) {
    printf("E: Loading land parcels from file: %s\n", fname);
  }
  file = [InFile create: [self getZone] withName: fname];
  if(file == nil) {
    fprintf(stderr, "%s -- could not open file ", sel_get_name(_cmd));
    perror(fname);
    return self;
  }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "EnvironmentXSize:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "EnvironmentXSize:", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getInt: &dummyXSize], fname, file)
    if(dummyXSize != [self getSizeX]) {
      fprintf(stderr, "%s -- wrong value for EnvironmentXSize in file %s. "
	      "Expecting %d found %d\n", sel_get_name(_cmd), fname,
	      [self getSizeX], dummyXSize);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "EnvironmentYSize:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "EnvironmentYSize:", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getInt: &dummyYSize], fname, file)
    if(dummyYSize != [self getSizeY]) {
      fprintf(stderr, "%s -- wrong value for EnvironmentYSize in file %s. "
	      "Expecting %d found %d\n", sel_get_name(_cmd), fname,
	      [self getSizeY], dummyYSize);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "LandParcelBitStringLength:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "LandParcelBitStringLength:",
	      word);
      [file drop];
      return self;
    }
  EOFCHECK([file getUnsigned: &dummyBSLength], fname, file)
    if(dummyBSLength != [parameter parcelBitStringSize]) {
      fprintf(stderr, "%s -- wrong value for LandParcelBitStringLength in file"
	      " %s. Expecting %u found %u\n", sel_get_name(_cmd), fname,
	      [parameter parcelBitStringSize], dummyBSLength);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "XCoord") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "XCoord", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "YCoord") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "YCoord", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "BitString") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "BitString", word);
      [file drop];
      return self;
    }
  buf = xmalloc([parameter parcelBitStringSize] + 2);
  // 2 is for '\0' at the end and one for luck -- which we'll need if the
  // bitstring size is 0
  if([arguments debugEnvironment]) {
    printf("E: New bit strings for land parcels:\n");
  }
  for(x = 0; x < [self getSizeX]; x++) {
    for(y = 0; y < [self getSizeY]; y++) {
      EOFCHECK([file getInt: &dummyX], fname, file)
	EOFCHECK([file getInt: &dummyY], fname, file)
	EOFCHECK([file getWord: buf], fname, file)
	lp = [self getObjectAtX: dummyX Y: dummyY];
      [[lp getAttributes] setToString: buf];
      if([arguments debugEnvironment]) {
	printf("\t%u at %d, %d): ", [lp getPIN], dummyX, dummyY);
	[[lp getAttributes] print];
      }
    }
  }
  [file drop];
  free(buf);
  return self;
}

/*

saveLandUsesToFileNamed:

Save the land uses to the file. A file called name.lu will be created if the
name doesn't habe a .lu suffix.

File format:

NumberOfLandUses: xxx
LandUseBitStringLength: nnn
LandUseIndex MatchBitString DontCareBitString
xx0		bbm0		bbdc0
xx1		bbm1		bbdc1
...

The index is the index in the array of land uses (not the land use PIN).

*/

-saveLandUsesToFileNamed: (char *)fname {
  id file;
  char *realfname;
  char *buf;
  int i;

  realfname = [MiscFunc getUsableFileNameFrom: fname withSuffix: ".lu"];
  if(realfname == NULL) {
    fprintf(stderr, "%s -- could not create datafile\n", sel_get_name(_cmd));
    return self;
  }
  file = [OutFile create: [self getZone] withName: realfname];
  if(file == nil) {
    fprintf(stderr, "%s -- could not create file ", sel_get_name(_cmd));
    free(realfname);
    perror(realfname);
    return self;
  }
  buf = xmalloc([parameter useBitStringSize] + 2);
  // 2 is for '\0' at the end and one for luck
  [file putString: "NumberOfLandUses:\t"];
  [file putUnsigned: [parameter nUses]];
  [file putNewLine];
  [file putString: "LandUseBitStringLength:\t"];
  [file putUnsigned: [parameter useBitStringSize]];
  [file putNewLine];
  [file putString: "LandUseIndex\tMatchBitString\tDontCareBitString\n"];
  for(i = 0; i < [parameter nUses]; i++) {
    [file putInt: i];
    [file putTab];
    [[[landUses atOffset: i] getMatchBitString] printToString: buf];
    [file putString: buf];
    [file putTab];
    [[[landUses atOffset: i] getDontCareBitString] printToString: buf];
    [file putString: buf];
    [file putNewLine];
  }
  [file drop];
  free(realfname);
  free(buf);
  return self;
}

/*

loadLandUsesFromFileNamed:

Load the land uses from the file. See saveLandUsesToFileNamed: for a
description of the format. As for loadLandParcelsFromFileNamed:, this assumes
that the land uses have already been created. And, again, there are numerous
horrible little checks to do for each stupid thing we load in.

*/

-loadLandUsesFromFileNamed: (char *)fname {
  id file;
  char *buf;
  char word[100];
  int i;
  unsigned dummyNUses, dummyBSLength, dummyI;
  LandUse *lu;

  if([arguments debugEnvironment]) {
    printf("E: Loading land uses from file: %s\n", fname);
  }
  file = [InFile create: [self getZone] withName: fname];
  if(file == nil) {
    fprintf(stderr, "%s -- could not open file ", sel_get_name(_cmd));
    perror(fname);
    return self;
  }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "NumberOfLandUses:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "NumberOfLandUses:", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getUnsigned: &dummyNUses], fname, file)
    if(dummyNUses != [parameter nUses]) {
      fprintf(stderr, "%s -- wrong value for NumberOfLandUses in file %s. "
	      "Expecting %u, found %u\n", sel_get_name(_cmd), fname,
	      [parameter nUses], dummyNUses);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "LandUseBitStringLength:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "LandUseBitStringLength:",
	      word);
      [file drop];
      return self;
    }
  EOFCHECK([file getUnsigned: &dummyBSLength], fname, file)
    if(dummyBSLength != [parameter useBitStringSize]) {
      fprintf(stderr, "%s -- wrong value for LandUseBitStringLength in file "
	      "%s. Expecting %u, found %u\n", sel_get_name(_cmd), fname,
	      [parameter useBitStringSize], dummyBSLength);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "LandUseIndex") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "LandUseIndex", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "MatchBitString") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "MatchBitString", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "DontCareBitString") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "DontCareBitString", word);
      [file drop];
      return self;
    }
  buf = xmalloc([parameter useBitStringSize] + 2);
  // 2 is for '\0' at the end and one for luck
  if([arguments debugEnvironment]) {
    printf("E: New bit strings for land uses:\n");
  }
  for(i = 0; i < (int)[parameter nUses]; i++) {
    EOFCHECK([file getInt: &dummyI], fname, file)
      lu = [landUses atOffset: i];
    EOFCHECK([file getWord: buf], fname, file)
      [[lu getMatchBitString] setToString: buf];
    EOFCHECK([file getWord: buf], fname, file)
      [[lu getDontCareBitString] setToString: buf];
    [lu updateParcelPortions];
    if([arguments debugEnvironment]) {
      printf("\t%u match: ", [lu getPIN]);
      [[lu getMatchBitString] printToFile: stdout];
      printf(" don't care: ");
      [[lu getDontCareBitString] print];
    }
  }
  [file drop];
  free(buf);
  return self;
}

/*

drop

Override the superclass method so that all items created by the environment
are also dropped.

*/

-(void)drop {
  [n_centre drop];
  [topology drop];
  [neighbourhood drop];
  [economy drop];
  [climate drop];
  [landUses deleteAll];
  [landParcels deleteAll];
  [super drop];
}

@end





