/*
    FEARLUS model0-5-1-5: HashTable.m
    Copyright (C) 1999-2004  Macaulay Institute

    This file is part of FEARLUS model0-5-1-5, an agent-based model of land use
    change.

    FEARLUS model0-5-1-5 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-5-1-5 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation of HashTable. Since you can only have one hash table at a time,
this is a set of class methods, with some static variables to keep track of
whether the table is in use (and a few extras).

*/

#import "HashTable.h"
#import <string.h>
#import <stdlib.h>
#import <errno.h>
#import <stdio.h>

static BOOL inuse = NO;		// Is the htable being used just now?
static size_t chunk;		// Estimated size of htable
static char **keys = NULL;	// Array to store all the keys (we have to do
				// this to free them anyway)
static id *values = NULL;	// Array to store the values (may as well, if
				// we're storing the keys...)
static int index;		// Next slot in the keys/values arrays

@implementation HashTable

/*

createHashEstimatedSize:

Create the hash table (call the hcreate function) with the estimated size of
hash table required. This method must be called before any of the others can
be used. Initialises some of the variables global to this class.

*/

+(BOOL)createHashEstimatedSize: (int)size {
  int retval;

  if(inuse) {
    fprintf(stderr, "HashTable: Error: Hash table is already in use\n");
    exit(1);
  }
  inuse = YES;
  retval = hcreate(size);
  if(retval == 0) {
    inuse = NO;
  }
  chunk = (size_t)size;
  index = 0;
  return inuse;
}

/*

addKey:data:

Call the hsearch(3C) function in ENTRY mode to add in a new item

*/

+(int)addKey: (char *)key data: (id)data {
  ENTRY item, *retval;
  char *mykey = strdup(key);

  if(!inuse) {
    fprintf(stderr, "HashTable: addKey:data: called with no initialisation\n");
    exit(1);
  }
  
  item.key = mykey;	
  item.data = (void *)data;

  retval = hsearch(item, ENTER);

  if(retval == NULL) {
    return -1;			// Hash table is full
  }
  else if(retval->data != item.data) {
    return 0;			// Entry already existed
  }
  else {
    if(index % chunk == 0) {
      keys = (char **)realloc(keys, (chunk * sizeof(char *)));
      values = (id *)realloc(values, (chunk * sizeof(id)));
      if(keys == NULL || values == NULL) {
	perror("HashTable: memory allocation");
	exit(1);
      }
    }
    keys[index] = mykey;
    values[index] = data;
    index++;
      
    return 1;			// Entry successfully added
  }
}

/*

getDataWithKey:

Return the data associated with a particular key.

*/

+(id)getDataWithKey: (char *)key {
  ENTRY item, *retval;

  if(!inuse) {
    fprintf(stderr,
	    "HashTable: getDataWithKey: called with no initialisation\n");
    exit(1);
  }

  item.key = key;
  retval = hsearch(item, FIND);
  if(retval == NULL) {
    return nil;
  }
  else {
    return (id)retval->data;
  }
}

/*

destroyHash

Destroy the hash table, free the keys, and reset the inuse flag to allow
another hash table to be created.

*/

+(void)destroyHash {
  int i;

  if(!inuse) {
    fprintf(stderr,
	    "HashTable: destroyHash called with no initialisation\n");
    exit(1);
  }

  hdestroy();
  for(i = 0; i < index; i++) {
    free(keys[i]);
  }
  free(keys);
  free(values);
  keys = NULL;
  values = NULL;
  inuse = NO;
  index = 0;
}

/*

getKeys:

We have to store the keys so we can free them when the table is destroyed, so
we may as well provide a method to allow objects to access them. Call with
a call-by-reference array -- returns the number of elements.

*/

+(int)getKeys: (char ***)keyptr {
  (*keyptr) = keys;
  return index;
}

/*

getValues:

Well, if we're keeping track of the keys, then why not the values? Again, arg
is call by ref, and the return value is the number of elements.

*/

+(int)getValues: (id **)valueptr {
  (*valueptr) = values;
  return index;
}

@end
