/*
    FEARLUS model0-5-1-5: ParcelCorrectedYieldWeightedCopyingStrategy.m
    Copyright (C) 1999-2004  Macaulay Institute

    This file is part of FEARLUS model0-5-1-5, an agent-based model of land use
    change.

    FEARLUS model0-5-1-5 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-5-1-5 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the ParcelCorrectedYieldWeightedCopyingStrategy object. 

*/

#import "ParcelCorrectedYieldWeightedCopyingStrategy.h"
#import "SelectUseBucket.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "LandManager.h"
#import "Environment.h"
#import "Parameter.h"
#import "FearlusArguments.h"
#import "BitString.h"

@implementation ParcelCorrectedYieldWeightedCopyingStrategy

/*

create:withManager:andParameters:

Create the strategy, setting the manager and parameters.

*/

+(id <ImitativeStrategy, HistoricalStrategy>)create: aZone
	   withManager: (LandManager *)lmgr
	 andParameters: (Parameter *)p {
  ParcelCorrectedYieldWeightedCopyingStrategy *obj;

  obj = [super create: aZone];
  obj->lm = lmgr;
  obj->parameter = p;
  return obj;
}

/*

decideLandUseForParcel:

Decide the land use for the land parcel, using a parcel corrected yield
copying strategy. The score for each parcel considered is yield * match,
with neighbourhood weighting for neighbouring land parcels.

*/

-(LandUse *)decideLandUseForParcel: (LandParcel *)lp {
  SelectUseBucket *bucket;
  id parcelIndex, parcelList, tempZone;
  LandParcel *myParcel, *neighbouringParcel;
  LandManager *neighbour;
  LandUse *lu;
  int i;
  double yield;
  double match;
  BitString *bs;

  tempZone = [Zone create: scratchZone];
  bucket = [SelectUseBucket create: tempZone
			    withParameters: parameter
			    andLandUses: [[lp getEnvironment] getLandUses]];
  // Loop through land parcels owned by the land manager
  if([arguments debugCoreMethods]) {
    printf("S: %s: Using parcel corrected yield weighted copying strategy of "
	   "land manager %u to determine land use for LP = land parcel %u at "
	   "(%d, %d)\n", sel_get_name(_cmd), [lm getPIN], [lp getPIN],
	   [lp getX], [lp getY]);
  }
  for(parcelIndex = [[lm getLandParcels] begin: tempZone], [parcelIndex next];
      [parcelIndex getLoc] == Member;
      [parcelIndex next]) {
    myParcel = [parcelIndex get];
    yield = (double)[myParcel getLastYield];
    match = (double)[ [lp getAttributes]
		      getMatchWith: [myParcel getAttributes] ];
    [bucket addScore: (yield * match)
	    toBucketForLandUse: [myParcel getLandUse]];
    if([arguments debugCoreMethods]) {
      printf("S: %s: Last yield of land parcel %u at (%d, %d) with match to LP"
	     " of %g and land use %u = %g. Score added = %g\n",
	     sel_get_name(_cmd), [myParcel getPIN], [myParcel getX],
	     [myParcel getY], match, [[myParcel getLandUse] getPIN], yield,
	     match * yield);
    }
  }
  [parcelIndex drop];
  // Loop through the neighbours, and through their land parcels
  bs = [BitString create: tempZone
		  setBitCount: [parameter parcelBitStringSize]];
  for(i = 0; neighbour = [lm getNextNeighbour: i]; i |= 1) {
    if([arguments debugCoreMethods]) {
      printf("S: %s: Adding scores for yield of land parcels belonging"
	     " to land manager %u\n", sel_get_name(_cmd), [neighbour getPIN]);
    }
    for(parcelIndex = [[neighbour getLandParcels] begin: tempZone],
	  [parcelIndex next];
	[parcelIndex getLoc] == Member;
	[parcelIndex next]) {
      neighbouringParcel = [parcelIndex get];
      [bs setToCopyOf: [neighbouringParcel getAttributes]];
      [bs mutateWithProbability: [lm getNeighbourNoise]];
      match = (double)[ [lp getAttributes] getMatchWith: bs ];
      yield = (double)[neighbouringParcel getLastYield];
      [bucket addScore: [lm getNeighbourWeight] * yield * match
	      toBucketForLandUse: [neighbouringParcel getLandUse]];
      if([arguments debugCoreMethods]) {
	printf("S: %s: Yield of land parcel %u at (%d, %d) with match to LP of"
	       " %g and use %u = %g. Score added = %g\n", sel_get_name(_cmd),
	       [neighbouringParcel getPIN], [neighbouringParcel getX],
	       [neighbouringParcel getY], match,
	       [[neighbouringParcel getLandUse] getPIN], yield,
	       yield * [lm getNeighbourWeight] * match);
      }
    }
    [parcelIndex drop];
  }
  lu = [bucket getChoice];
  [bucket drop];
  [bs drop];
  [tempZone drop];
  return lu;
}

@end
