/*
    FEARLUS model0-5-1-5: YieldWeightedTemporalCopyingStrategy.m
    Copyright (C) 1999-2004  Macaulay Institute

    This file is part of FEARLUS model0-5-1-5, an agent-based model of land use
    change.

    FEARLUS model0-5-1-5 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-5-1-5 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the YieldWeightedTemporalCopyingStrategy object. 

*/

#import "YieldWeightedTemporalCopyingStrategy.h"
#import "SelectUseBucket.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "LandManager.h"
#import "Environment.h"
#import "Parameter.h"
#import "FearlusArguments.h"

@implementation YieldWeightedTemporalCopyingStrategy

/*

create:withManager:andParameters:

Create the strategy, setting the manager and parameters.

*/

+(id <ImitativeStrategy, HistoricalStrategy>)create: aZone
	   withManager: (LandManager *)lmgr
	 andParameters: (Parameter *)p {
  YieldWeightedTemporalCopyingStrategy *obj;

  obj = [super create: aZone];
  obj->lm = lmgr;
  obj->parameter = p;
  return obj;
}

/*

decideLandUseForParcel:

Decide the land use for the land parcel, using a simple copying strategy.
Create a bucket, which is filled with y * neighbourhood weighting of the
land manager for each land use used by a parcel owned by a neighbour of the
land manager, and y for each land use used by a parcel owned by the land
manager, where y is the yield of the potentially copied land parcel.

*/

-(LandUse *)decideLandUseForParcel: (LandParcel *)lp {
  SelectUseBucket *bucket;
  id parcelIndex, parcelList, tempZone;
  LandParcel *myParcel, *neighbouringParcel;
  LandManager *neighbour;
  LandUse *lu;
  int i, y;
  double yield;

  tempZone = [Zone create: scratchZone];
  bucket = [SelectUseBucket create: tempZone
			    withParameters: parameter
			    andLandUses: [[lp getEnvironment] getLandUses]];
  // Loop through land parcels owned by the land manager
  if([arguments debugCoreMethods]) {
    printf("S: %s: Using yield weighted copying strategy of land manager"
	   " %u to determine land use for land parcel %u at (%d, %d)\n",
	   sel_get_name(_cmd), [lm getPIN], [lp getPIN], [lp getX], [lp getY]);
  }
  for(y = 0; y < [lm getMemorySize]; y++) {
    for(parcelIndex = [[lm getLandParcels] begin: tempZone],
	  [parcelIndex next];
	[parcelIndex getLoc] == Member;
	[parcelIndex next]) {
      myParcel = [parcelIndex get];
      lu = [myParcel recallLandUseAgo: (unsigned)y];
      if(lu == nil) continue;
      yield = (double)[myParcel recallYieldAgo: (unsigned)y];
      [bucket addScore: yield toBucketForLandUse: lu];
      if([arguments debugInnerCoreMethods]) {
	printf("S: %s: Yield of land parcel %u at (%d, %d) with land use"
	       " %u = %g (%d years ago)\n", sel_get_name(_cmd),
	       [myParcel getPIN], [myParcel getX], [myParcel getY],
	       [lu getPIN], yield, y);
      }
    }
    [parcelIndex drop];
    // Loop through the neighbours, and through their land parcels
    for(i = 0; neighbour = [lm getNextNeighbour: i]; i |= 1) {
      if([arguments debugCoreMethods]) {
	printf("S: %s: Adding scores for yield of land parcels belonging"
	       " to land manager %u\n", sel_get_name(_cmd),
	       [neighbour getPIN]);
      }
      for(parcelIndex = [[neighbour getLandParcels] begin: tempZone],
	    [parcelIndex next];
	  [parcelIndex getLoc] == Member;
	  [parcelIndex next]) {
	neighbouringParcel = [parcelIndex get];
	lu = [lm recallLandUseForParcel: neighbouringParcel
		 yearsAgo: (unsigned)y];
	if(lu == nil) continue;
	yield = (double)[lm recallYieldForParcel: neighbouringParcel
			    yearsAgo: (unsigned)y];
	[bucket addScore: [lm getNeighbourWeight] * yield
		toBucketForLandUse: lu];
	if([arguments debugInnerCoreMethods]) {
	  printf("S: %s: Yield of land parcel %u at (%d, %d) with land use "
		 "%u = %g (%d years ago). Score of land use incremented by "
		 "%g\n", sel_get_name(_cmd), [neighbouringParcel getPIN],
		 [neighbouringParcel getX], [neighbouringParcel getY],
		 [lu getPIN], yield, y, yield * [lm getNeighbourWeight]);
	}
      }
      [parcelIndex drop];
    }
  }
  lu = [bucket getChoice];
  [bucket drop];
  [tempZone drop];
  return lu;
}

@end
